import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import log_analysis_util

PKT_TYPE_LTG = 3
PKT_TYPE_MGMT = 11

T_MIN = 0E6
T_MAX = 300E6

#Define nodes - each entry must be 2-tuple:
# (node_name, HDF5_raw_log_file_name)
nodes = (
    ('IBSS 1',    'E0_R18_n_ibss_1.hdf5'),
    ('IBSS 2',    'E0_R18_n_ibss_2.hdf5'),
    ('IBSS 3',    'E0_R18_n_ibss_3.hdf5'),
)

flow_conds_common = (
    ('timestamp', '>', T_MIN),
    ('timestamp', '<', T_MAX),
)

flow_cond_ltg = ('pkt_type', '==', PKT_TYPE_LTG)

#Define flows - each entry must be 4-tuple:
# (src, dest, conds_tx, conds_rx)
# src/dest: either string name matching name in nodes list or MAC address as uint64
# conds: list of 3-tuples (field_name, operator, value); use [] for no conditions
flows = (
    ('IBSS 1',    'IBSS 2', [flow_cond_ltg], [flow_cond_ltg]),
    ('IBSS 1',    'IBSS 3', [flow_cond_ltg], [flow_cond_ltg]),
    ('IBSS 2',    'IBSS 1', [flow_cond_ltg], [flow_cond_ltg]),
    ('IBSS 2',    'IBSS 3', [flow_cond_ltg], [flow_cond_ltg]),
    ('IBSS 3',    'IBSS 1', [flow_cond_ltg], [flow_cond_ltg]),
    ('IBSS 3',    'IBSS 2', [flow_cond_ltg], [flow_cond_ltg])
)

#Plotting colors per flow - each entry must be 3-tuple:
# (color_Tx, color_Rx_good, color_Rx_bad)
# Each color is 4-bytes as u32: (alpha, red, green, blue)
flow_colors = (
    (0xFFFF0000, 0xA0FF0000, 0xA0AA0000), #Red
    (0xFF00CC00, 0xA000CC00, 0xA0007700), #Green
    (0xFFFF9500, 0xA0FF9500, 0xA0AA4000), #Orange
    (0xFF0C5DA5, 0xA00C5DA5, 0xA0000850), #Blue
    (0xFFFF33CC, 0xFFFF33CC, 0xFFFF33CC), #Magenta
    (0xFFCCCC52, 0xFFCCCC52, 0xFFCCCC52), #Yellowish
    (0xFF000000, 0xFF000000, 0xFF222222), #Black
)

#%%

#Extract Tx/Rx entries per flow per node
flows_tx, flows_rx, all_entries_np = log_analysis_util.proc_txrx_logs_to_flows(nodes, flows, flow_conds_common)

#%%


##########################################
# Plot #1: Rx throughput vs. time per flow
#  Process:
#   1) Extract times and Rx lengths from destination node of each flow
#   2) Resample Rx lengths with regular interval
#   3) Merge time indexes of resampled Rx length vectors
#   4) Compute rolling mean of each resampled Rx length vector
#   5) Scale length vs. time vector to be throughput in Mbps
#   6) Plot individual flow throughputs and their sum

#Resample interval (miliseconds)
rs_interval = 1

#Rolling window (# of samples per window)
rolling_window = 1000

#Itereate over flows to extract and resample Rx length vectors from destination 
# node for each flow 
flows_rx_lens_rs = list()
for f_idx,flow in enumerate(flows):

    f_src = flow[0]
    f_dst = flow[1]

    #Extract all Rx entries from target node
    f_rx = flows_rx[f_idx][ [n[0] for n in nodes].index(f_dst) ]

    #Select good-checksum, non-duplicated receptions
    f_rx_idx = ((f_rx['flags'] & 0x1) == 0) & (f_rx['fcs_result'] == 0)
    f_rx = f_rx[f_rx_idx]
    
    f_rx_t    = f_rx['timestamp']
    f_rx_lens = f_rx['length']

    #Insert leading and trailing zeros, so the rolling window below
    # has valid data on both sides of the actual data
    f_rx_t    = np.insert(f_rx_t, 0, 0)
    f_rx_t    = np.insert(f_rx_t, len(f_rx_t)-1, np.max(f_rx_t)+(rolling_window*rs_interval*1000))

    f_rx_lens = np.insert(f_rx_lens, 0, 0)
    f_rx_lens    = np.insert(f_rx_lens, len(f_rx_lens)-1, 0)
    
    #Convert Rx timestamps to Pandas datetime index
    f_rx_t_pd = pd.to_datetime(f_rx_t, unit='us')

    #Convert Rx pkt length array to Pandas series indexed by Rx timestamps
    f_rx_lens_pd = pd.Series(f_rx_lens, index=f_rx_t_pd)  
    
    #Resample Rx length series to regular sampling interval
    f_rx_lens_pd_rs = f_rx_lens_pd.resample(('%dL' % rs_interval), how='sum').fillna(value=0)

    #Remember this flow's resampled Rx lengths for later plotting
    flows_rx_lens_rs.insert(f_idx, f_rx_lens_pd_rs)

    #Construct the running union of all flow's resampled Rx timestamps
    if(f_idx == 0):
        flows_rx_t_idx = f_rx_lens_pd_rs.index
    else:
        flows_rx_t_idx = flows_rx_t_idx.union(f_rx_lens_pd_rs.index)

#Itereave over resampled Rx length vectors to merge indexes and calculate mean throughput
flow_xputs = list()
for f_idx in range(len(flows_rx_lens_rs)):
    #Re-index each flow's Rx lengths series by the common-to-all-flows time index
    flows_rx_lens_rs[f_idx] = flows_rx_lens_rs[f_idx].reindex(flows_rx_t_idx)

    #Compute the rolling mean of the Rx lengths
    f_xput = pd.rolling_mean(flows_rx_lens_rs[f_idx], window=rolling_window, min_periods=1).fillna(value=0)

    #Scale the Rx lengths to throughputs
    # 1.0e-6: Convert microseconds to seconds
    # 8.0: Convert bytes to bits
    # (1.0/(rs_interval * 1e-3): Convert resampling interval in msec to seconds
    f_xput = f_xput  * (1.0e-6 * 8.0 * (1.0/(rs_interval * 1e-3)))

    #Remember each flow's xput for plotting
    flow_xputs.insert(f_idx, f_xput)


#Create time vector for plot X-axis, in units of seconds
t_sec = flows_rx_t_idx.astype('int64') / 1.0E9
plt_t = t_sec

#Open a clean figure for plotting
plt.close(1)
fig = plt.figure(1, figsize=(10,8))

#Plot each throughput vector and their sum
sum_xput = flow_xputs[0]
for f_idx,f_xput in enumerate(flow_xputs):
    if(f_idx > 0):
        sum_xput += f_xput

    plt.plot(plt_t, f_xput, 
        color='#%06x' % (0xFFFFFF & flow_colors[f_idx][0]), #Color as HTML
        label='Flow {0}'.format(f_idx+1))

plt.plot(plt_t, sum_xput, color='#000000', label='Sum of Flows')

#Place a legend above the axes
leg = plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=8,
       ncol=4, borderaxespad=0.)
       
#Make the legend line widths thicker
for legobj in leg.legendHandles:
    legobj.set_linewidth(5.0)


#Label and format the axes
plt.xlabel('Time (sec)')
plt.ylabel('Throughput (Mb/sec)')
plt.xlim([(T_MIN/1E6) + 5, (T_MAX/1E6) - 5])
plt.grid('on')

plt.savefig('xput_vs_time.png')

#%%
############################################
# Plot #2: Rx throughput histograms per flow
# Uses same data as Plot #1, plotted as PDFs
#Open a clean figure for plotting

#Select data with time values between 0+3s and END-3s
# This drops the ramping up/down values from the rolling window, which aren't
#  actual throughputs, just analysis artifacts
tsel = (((flows_rx_t_idx.astype('int64') / 1.0E3) > T_MIN+3E6) & 
        ((flows_rx_t_idx.astype('int64') / 1.0E3) < (T_MAX-3E6)))

#matplotlib hist norm vs weights is weird:
# http://stackoverflow.com/questions/5498008/pylab-histdata-normed-1-normalization-seems-to-work-incorrect
hist_weights = np.ones((sum(tsel),))/sum(tsel)

hist_bins = 25

hist_lims = [0,12,0,0.25]

plt.close(2)
fig = plt.figure(2, figsize=(10,18))

for f_idx,f_xput in enumerate(flow_xputs):
    plt.subplot(6,1,f_idx+1)
    f_xput_sel = f_xput[tsel].values
    plt.hist(f_xput_sel, 
             bins=hist_bins,
             weights=hist_weights, 
             histtype='stepfilled', 
             color='#%06x' % (0xFFFFFF & flow_colors[f_idx][0]), #Color as HTML
             label='Flow {0}'.format(f_idx))
    plt.vlines(np.mean(f_xput_sel), hist_lims[2], hist_lims[3], linewidth=2.0)
    plt.axis(hist_lims); plt.grid('on')
    plt.ylabel('Probability')
    plt.title('Flow {0} Throughput Probability Density Function'.format(f_idx+1))
plt.xlabel('Throughput (Mb/sec)')

plt.savefig('xput_hist.png')
#%%
#######################################
# Plot #3: Rx power vs. time
#  Process:
#   1) Extract times and Rx powers from each node's log, grouped by transmitting node
#   2) Resample Rx powers with regular interval
#   3) Compute rolling mean of each resampled Rx power vector
#   4) Plot each inter-node power vector

#Extract all Rx entries from every node log, grouped by node
all_rx   = [log['RX_OFDM'] for log in all_entries_np]

#Extract MAC addresses for each node
node_addrs = [log['NODE_INFO']['wlan_mac_addr'][0] for log in all_entries_np]

#Resample interval (miliseconds)
rs_interval = 1

#Rolling window (# of samples per window)
rolling_window = 1000

#Open a clean figure for plotting
plt.close(3)
fig = plt.figure(3, figsize=(10,8))

for idx_nRx,node_Rx in enumerate(nodes):
    rx_log = all_rx[idx_nRx]
  
    for idx_nTx,node_Tx in enumerate(nodes):
        if(idx_nRx == idx_nTx):
            #Skip Rx for same Tx
            continue
    
        addr_tx = node_addrs[idx_nTx]
    
        #Find all Rx packet from this transmitter's MAC address
        rx_from_txNode = rx_log[ rx_log['addr2'] == addr_tx]
    
        rx_t    = rx_from_txNode['timestamp']
        rx_pwrs = rx_from_txNode['power']
        
        #Convert Rx timestamps to Pandas datetime index
        rx_t_pd = pd.to_datetime(rx_t, unit='us')
    
        #Convert Rx length/power arrays to Pandas series indexed by Rx timestamps
        rx_pwrs_pd = pd.Series(rx_pwrs, index=rx_t_pd)  
        
        #Resample Rx length/power series to regular sampling interval
        rx_pwrs_pd_rs = rx_pwrs_pd.resample(('%dL' % rs_interval), how='mean').fillna(method='ffill')
        rx_pwrs_pd_mean = pd.rolling_mean(rx_pwrs_pd_rs, window=rolling_window, min_periods=1).fillna(value=0)
    
        #Plot averaged Rx power curve vs. time (converted to seconds)
        t_plt = rx_pwrs_pd_mean.index.astype('int64')/1.0E9
        plt.plot(t_plt, rx_pwrs_pd_mean, label='{0}->{1}'.format(node_Tx[0], node_Rx[0]))


leg = plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=8,
       ncol=3, borderaxespad=0.)
       
for legobj in leg.legendHandles:
    legobj.set_linewidth(5.0)
       
plt.xlabel('Time (sec)')
plt.ylabel('Rx Power (dBm)')

plt.xlim([(T_MIN/1E6) + 5, (T_MAX/1E6) - 5])
plt.ylim([-90,-20])
plt.grid('on')

plt.savefig('rxPow_vs_time.png')

#%%
##################################################
# Plot #4: Contention window vs. time at each node
#  Process:
#   1) Extract times and DCF contention window values from each node's log
#   2) Resample contention window samples with regular interval
#   3) Compute rolling mean of each resampled vector
#   4) Plot each node's averaged CW vector

#Extract all Tx_Low entries from every node log
all_tx   = [log['TX_LOW'] for log in all_entries_np]

#Open a clean figure for plotting
plt.close(4)
fig = plt.figure(4, figsize=(10,8))

for idx_nTx,node_Tx in enumerate(nodes):

    #Extract all CW values from Tx log
    tx_cw = all_tx[idx_nTx]['cw']
    tx_t  = all_tx[idx_nTx]['timestamp']

    #Convert Rx timestamps to Pandas datetime index
    tx_t_pd = pd.to_datetime(tx_t, unit='us')

    #Convert Rx length/power arrays to Pandas series indexed by Rx timestamps
    tx_cw_pd = pd.Series(tx_cw, index=tx_t_pd)  

    #Resample CW samples to regular interval
    tx_cw_pd_rs = tx_cw_pd.resample(('%dL' % rs_interval), how='mean').fillna(method='ffill')

    #Take rolling mean of resampled vector
    tx_cw_pd_mean = pd.rolling_mean(tx_cw_pd_rs, window=rolling_window, min_periods=1).fillna(value=0)

    #Plot averaged Rx power curve vs. time (converted to seconds)
    t_plt = tx_cw_pd_mean.index.astype('int64')/1.0E9
    plt.semilogy(t_plt, tx_cw_pd_mean, label='{0}'.format(node_Tx[0]))

leg = plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=8,
       ncol=3, borderaxespad=0.)
       
for legobj in leg.legendHandles:
    legobj.set_linewidth(5.0)

       
plt.xlabel('Time (sec)')
plt.ylabel('Contention Window Size')

# Set the Y ticks and tick labels
cw_yticks = [1,3,7,15,31,63,127,255,511,1023]
plt.yticks( cw_yticks, map(str, cw_yticks) )
plt.ylim( [14, 1100] )

plt.xlim([(T_MIN/1E6) + 5, (T_MAX/1E6) - 5])
plt.grid('on')

plt.savefig('cw_vs_time.png')
